import os,sys,time
from sklearn.cluster import spectral_clustering
import numpy as np 
from sklearn.preprocessing import normalize
sys.path.append('/home/yinshuo/usr/xmipp/lib')
import xmipp

def correntropy(x, y):
    return np.average(np.exp(-1 * np.power(x - y, 2)))

def correlation(img):
    N = img.shape[0]
    img = np.reshape(img, [N, -1])
    img = normalize(img)
    meaned = np.mean(img, axis=0)
    img = img - meaned
    corr = np.dot(img, img.T)
    return corr

def nearestK(data, k):
    graph = np.zeros((data.shape[0], data.shape[0]))
    for row in range(data.shape[0]):
        dis = [complex(-1 * data[row, i], i) for i in range(data.shape[1])]
        sortedDis = np.sort_complex(dis)
        firstk = sortedDis[1:1+k].imag
        for index in firstk:
            graph[row, int(index)] = 1
    return graph

def mutualKNN(data, k):
    graph = nearestK(data, k)
    return np.asarray(graph*graph.T)

def Jaccard(adj):
    N = adj.shape[0]
    d = np.dot(adj, np.ones(N))
    D = np.broadcast_to(d, (N,N))
    intersection = np.dot(adj,adj)
    union = D+D.T-intersection+1e-5
    return intersection/union

def Sorensen(adj):
    N = adj.shape[0]
    d = np.dot(adj, np.ones(N))
    D = np.broadcast_to(d, (N,N))
    intersection = np.dot(adj,adj)
    union = D+D.T+1e-5
    return intersection/union

def HD(adj):
    N = adj.shape[0]
    d = np.dot(adj, np.ones(N))
    D = np.broadcast_to(d, (N,N))
    intersection = np.dot(adj,adj)
    union = np.maximum(D,D.T)+1e-5
    return intersection/union

def RA(adj):
    N = adj.shape[0]
    d = 1.0 / np.sqrt(np.dot(adj, np.ones(N)))
    ra = adj*d
    ra = np.dot(ra,ra.T)
    m = np.max(ra) * 1.1
    for i in range(N):
        ra[i,i] = m
    return ra

def Katz(adj, b):
    (v, w) = np.linalg.eig(adj)
    beta = 1 / np.max(v) / b
    katz = np.linalg.inv(np.eye(v.size) - beta * adj) - np.eye(v.size)
    m = np.max(katz)
    for i in range(katz.shape[0]):
        katz[i, i] = m*1.1
    return katz

def LHN2(adj, phi):
    (v, w) = np.linalg.eig(adj)
    lambda1 = np.max(v)
    d = np.dot(adj, np.ones(adj.shape[0]))
    DI = np.diag(1.0/d)
    T = np.linalg.inv(np.eye(adj.shape[0]) - phi * adj / lambda1)
    return np.dot(np.dot(DI, T), DI)

def align_images(img_path):
    cmd = 'xmipp_image_align -i %s --oroot align --do_not_check_mirrors'%(img_path)
    os.system(cmd)
    cmd = 'xmipp_transform_geometry -i align_alignment.xmd -o align.stk --apply_transform'
    os.system(cmd)

def getcorr(name='align.stk'):
    img = xmipp.Image()
    img.read(name)
    data = img.getData()
    return correlation(data)

# def getadj(K):
#     corr = getcorr()
#     adj = mutualKNN(corr, K)
#     img = xmipp.Image()
#     img.setData(corr)
#     img.write('corr.stk')
#     img.setData(adj)
#     img.write('adj.stk')
#     return adj

def getadj(threads, K):
    cmd = 'mpirun -np %d main ./align.xmd %d'%(threads, K)
    os.system(cmd)
    img = xmipp.Image()
    img.read('adj.stk')
    return img.getData()

def divisiveSperctral(adj, kn):
    cluster = spectral_clustering(adj, 2)
    # cluster = spectral_cl(adj)
    k = 2
    while k < kn:
        count = [np.sum(cluster == i) for i in range(k)]
        largest_cluster = np.argmax(count)
        nodes = [i for i in range(cluster.size) if cluster[
            i] == largest_cluster]
        sub_graph = adj[nodes, :][:, nodes]
        sub_cluster = spectral_clustering(sub_graph, 2)
        for i in range(len(nodes)):
            if(sub_cluster[i] == 1):
                cluster[nodes[i]] = k
        k += 1
    return cluster

def divisiveSpectral(stkfilename, adj, Kn, fun, use_cor, *args):
    img = xmipp.Image()
    img.read(stkfilename)
    data = img.getData()
    avg = np.mean(data, axis=(0,1))
    adjs = [adj]
    avgs = [avg]
    cluster = np.zeros(adj.shape[0], dtype='int32')
    k=1
    while k < Kn:
        count = [np.sum(cluster == i) for i in range(k)]
        largest_cluster = np.argmax(count)
        nodes = np.where(cluster==largest_cluster)[0]
        sub_graph = adjs[largest_cluster]
        sub_cluster = spectral_clustering(sub_graph, 2)
        for i in range(len(nodes)):
            if(sub_cluster[i] == 1):
                cluster[nodes[i]] = k
        sub = np.where(cluster==largest_cluster)[0]
        update_cluster(stkfilename, sub, avgs, adjs, largest_cluster, fun, use_cor, *args)
        adjs.append(None)
        avgs.append(None)
        sub = np.where(cluster==k)[0]
        update_cluster(stkfilename, sub, avgs, adjs, k, fun, use_cor, *args)
        k+=1
    img.setData(np.array(avgs))
    img.write('new_%s_centers.stk'%(fun.__name__))
    return cluster

def update_cluster(stkfilename, cluster, avgs, adjs, pos, fun, use_cor, *args):
    xmd = xmipp.MetaData()
    xmd.addLabel(xmipp.MDL_IMAGE)
    print(len(cluster))
    for idx in cluster:
        i = xmd.addObject()
        xmd.setValue(xmipp.MDL_IMAGE, str(idx+1)+'@'+stkfilename, i)
    xmd.write('tmpstk.xmd')
    cmd = 'xmipp_image_align -i tmpstk.xmd --oroot tmp --do_not_check_mirrors'
    os.system(cmd)
    cmd = 'xmipp_transform_geometry -i tmp_alignment.xmd -o tmp.stk --apply_transform'
    os.system(cmd)
    tmp = xmipp.Image()
    tmp.read('tmp.stk')
    avgs[pos] = np.mean(tmp.getData(), axis=(0,1))
    adjs[pos] = construct_adj('tmp.xmd', max(len(cluster)/2, 10), fun, use_cor, *args)

def construct_adj(xmdfilename, K, fun, use_cor, *args):
    if not use_cor:
        cmd = 'mpirun -np %d main ./%s %d'%(4, xmdfilename, K)
        os.system(cmd)
        img = xmipp.Image()
        img.read('adj.stk')
        adj = img.getData()
    else:
        corr = getcorr('tmp.stk')
        adj = mutualKNN(corr, K)
    if len(args)>0:
        corr = fun(adj, args[0])
    else:
        corr = fun(adj)
    mknn = mutualKNN(corr, K)
    return mknn


def clustering(stkfilename,adj, fun, K, Kn, use_cor, *args):
    start = time.time()
    if len(args)>0:
        corr = fun(adj, args[0])
    else:
        corr = fun(adj)
    mknn = mutualKNN(corr, K)
    img = xmipp.Image()
    img.setData(mknn)
    img.write(fun.__name__+'adj.stk')
    cluster = divisiveSpectral(stkfilename, mknn, Kn, fun, use_cor, *args)
    getCenters(cluster, stkfilename, fun.__name__)
    end = time.time()
    print '%s cost %d seconds'%(fun.__name__, end-start)
    return cluster

def getCenters(clusters, stkfilename, prefix):
    # some info
    classes = int(np.max(clusters)) + 1
    img = xmipp.Image()
    img.read('1@' + stkfilename)
    shape = img.getData().shape
    xmdname = prefix + '_classes.xmd'
    stkname = prefix + "_centers.stk"
    dirname = os.path.dirname(stkfilename) + '/'
    # some initailize
    centers = np.zeros((classes, 1, shape[0], shape[1]))
    count = [0] * classes
    mdClass = xmipp.MetaData()
    mdClass.addLabel(xmipp.MDL_IMAGE)
    mdClass.addLabel(xmipp.MDL_CLASS_COUNT)
    mdList = []
    for i in range(classes):
        tmp = xmipp.MetaData()
        tmp.addLabel(xmipp.MDL_IMAGE)
        mdList.append(tmp)

    for i in range(clusters.size):
        path = str(i + 1) + '@' + stkfilename
        img.read(path)
        centers[clusters[i], 0] += img.getData()
        count[clusters[i]] += 1
        mdList[clusters[i]].addObject()
        mdList[clusters[i]].setValue(xmipp.MDL_IMAGE, path, count[clusters[i]])

    for i in range(classes):
        blockname = 'class%06d_images'%(i+1)
        mdList[0].write(blockname+'@'+ dirname + xmdname, xmipp.MD_APPEND)
        del mdList[0]
        centers[i, 0] /= count[i]
        centerPath = str(i + 1) + '@' + dirname + stkname
        mdClass.addObject()
        mdClass.setValue(xmipp.MDL_IMAGE, centerPath, i + 1)
        mdClass.setValue(xmipp.MDL_CLASS_COUNT, long(count[i]), i + 1)

    centerStk = xmipp.Image()
    centerStk.setData(centers)
    # centerStk.write(dirname+stkname)
    mdClass.write('classes@' + dirname + xmdname, xmipp.MD_APPEND)
    return centers
